from typing import ClassVar
from typing import Protocol

from gyver.database.types import Driver


class Dialect(Protocol):
    default_port: ClassVar[int]
    driver: ClassVar[Driver]
    dialect_name: ClassVar[str]
    async_driver: ClassVar[str]
    sync_driver: ClassVar[str]
    only_host: ClassVar[bool]



from contextlib import suppress
import logging
from typing import BinaryIO, Optional, Protocol, cast

from botocore.exceptions import ClientError

from gyver import utils
from gyver.exc import GyverError
from gyver.config import ProviderConfig
from gyver.config import from_config

from ._client import make_client


class StorageConfig(ProviderConfig):
    _prefix_ = "storage"

    access_key_id: str
    secret_access_key: str
    bucket_name: str
    region_name: str
    presigned_expiration: int = 3600
    endpoint_url: Optional[str] = None


class FileNotFound(GyverError, FileNotFoundError):
    pass


class ReadableBuffer(Protocol):
    def read(self) -> bytes:
        ...


class StorageProvider:
    def __init__(self, config: Optional[StorageConfig] = None) -> None:
        self._config = config or from_config(StorageConfig)

    @utils.lazyfield
    def client(self):
        kwargs = {
            "region_name": self._config.region_name,
            "aws_access_key_id": self._config.access_key_id,
            "aws_secret_access_key": self._config.secret_access_key,
        }
        if endpoint_url := self._config.endpoint_url:
            kwargs["endpoint_url"] = endpoint_url
        return make_client("s3", **kwargs)

    @property
    def config(self):
        return self._config

    def _generate_presigned_url(
        self,
        method: str,
        object_name: str,
        bucket_name: Optional[str] = None,
    ):
        bucket_name = bucket_name or self._config.bucket_name
        try:
            response = self.client.generate_presigned_url(
                method,
                Params={"Bucket": bucket_name, "Key": object_name},
                ExpiresIn=self._config.presigned_expiration,
            )
        except ClientError as e:
            logging.error(e)
            return object_name
        else:
            return response

    def upload_file(
        self,
        buf: BinaryIO,
        object_name: str,
        bucket_name: Optional[str] = None,
    ):
        bucket_name = bucket_name or self._config.bucket_name
        self.client.upload_fileobj(buf, bucket_name, object_name)

    def generate_presigned_get(
        self, object_name: str, bucket_name: Optional[str] = None
    ):
        bucket_name = bucket_name or self._config.bucket_name
        return (
            self._generate_presigned_url(
                "get_object", object_name, bucket_name
            )
            if self.is_key(object_name, bucket_name)
            else object_name
        )

    def is_key(
        self, object_name: str, bucket_name: Optional[str] = None
    ) -> bool:
        bucket_name = bucket_name or self._config.bucket_name
        try:
            self.client.head_object(Bucket=bucket_name, Key=object_name)
        except ClientError:
            return False
        else:
            return True

    def get_file(
        self, object_name: str, bucket_name: Optional[str] = None
    ) -> ReadableBuffer:
        bucket_name = bucket_name or self._config.bucket_name
        if not self.is_key(object_name, bucket_name):
            raise FileNotFound(
                f"Object {object_name} not found in {bucket_name} bucket"
            )
        response = self.client.get_object(Bucket=bucket_name, Key=object_name)
        return response["Body"]



from typing import Optional
from typing import Sequence

from gyver import utils
from gyver.config import ProviderConfig

from . import drivers
from .types import Driver


class DatabaseConfig(ProviderConfig):
    driver: Driver
    host: str
    port: int = -1
    user: str = ""
    password: str = ""
    name: str = ""

    @utils.lazyfield
    def real_port(self) -> int:
        return self.port if self.port != -1 else self.dialect.default_port

    @utils.lazyfield
    def dialect(self) -> drivers.Dialect:
        return drivers.resolve_driver(self.driver)


def make_database_config(
    prefix: str,
    without_prefix: Sequence[str] = (),
    name: str = "DatabaseConfig",
    module: Optional[str] = __name__,
) -> type[DatabaseConfig]:
    return type(
        name,
        (DatabaseConfig,),
        {
            "__prefix__": prefix,
            "__without_prefix__": without_prefix,
            "__module__": module,
        },
    )


DefaultDatabaseConfig = make_database_config("db")



from pydantic import BaseModel

from gyver import utils


class Model(BaseModel):
    class Config:
        json_loads = utils.json.loads
        json_dumps = utils.json.dumps
        frozen = True
        orm_mode = True
        alias_generator = utils.to_camel
        allow_population_by_field_name = True
        keep_untouched = (utils.lazyfield,)


class MutableModel(BaseModel):
    class Config:
        frozen = False



from typing import TYPE_CHECKING
from typing import Any
from typing import Literal

from boto3 import Session

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client


def make_client(service_name: Literal["s3"], **params: Any) -> "S3Client":
    return Session().client(service_name, **params)



import typing

T = typing.TypeVar("T")
SelfT = typing.TypeVar("SelfT")


class lazyfield(typing.Generic[SelfT, T]):
    def __init__(self, func: typing.Callable[[SelfT], T]) -> None:
        self._func = func

    def __set_name__(self, owner: type[SelfT], name: str):
        self.public_name = name
        self.private_name = f"_lazyfield_{name}"

    def __get__(
        self,
        instance: typing.Optional[SelfT],
        owner: typing.Optional[type[SelfT]] = None,
    ) -> T:
        if not instance:
            assert owner is not None
            raise AttributeError(
                f"{owner.__name__!r} has no" f"attribute {self.public_name!r}"
            )
        assert instance is not None
        try:
            val = typing.cast(
                T,
                object.__getattribute__(
                    instance,
                    self.private_name,
                ),
            )
        except AttributeError:
            val = self._try_set(instance)
        return val

    def _try_set(self, instance: SelfT) -> T:
        try:
            val = self._func(instance)
        except Exception as e:
            # remove exception context to create easier traceback
            raise e from None
        else:
            object.__setattr__(instance, self.private_name, val)
            return val

    def __set__(self, instance: SelfT, value: T):
        setattr(instance, self.private_name, value)

    def __delete__(self, instance: SelfT):
        delattr(instance, self.private_name)



class GyverError(Exception):
    """Base exception for all gyver exceptions"""


class InvalidCast(GyverError):
    """Exception raised for config when
    cast callable raises an error"""


class MissingName(GyverError, KeyError):
    """Exception raised for config when
    name is not found in given environ"""



import typing
from dataclasses import dataclass

import sqlalchemy as sa

from . import _helpers
from . import comp as cp
from . import interface
from .interface import BindClause
from .interface import Mapper

T = typing.TypeVar("T")


@dataclass
class Where(typing.Generic[T]):
    field: str
    expected: typing.Optional[T] = None
    comp: interface.Comparator[T] = cp.equals

    def bind(self, entity: Mapper) -> interface.Comparison:
        attr = _helpers.retrieve_attr(entity, self.field)
        return (
            sa.true()
            if self.expected is None
            else self.comp(attr, self.expected)
        )


@dataclass(frozen=True)
class _JoinBind:
    items: typing.Sequence[BindClause]
    operator: typing.Callable[..., interface.Comparison]

    def bind(self, entity: Mapper) -> interface.Comparison:
        return self.operator(*(item.bind(entity) for item in self.items))


def and_(*bind: BindClause) -> _JoinBind:
    return _JoinBind(bind, sa.and_)


def or_(*bind: BindClause) -> _JoinBind:
    return _JoinBind(bind, sa.or_)
